/****************************************************************************
*                                                                           *
*                  ___  ____  ____   ___  __    ____  ____                  *
*                 / __)(_  _)(  _ \ / __)(  )  ( ___)(  _ \                 *
*                ( (__  _)(_  )   /( (__  )(__  )__)  )(_) )                *
*                 \___)(____)(_)\_) \___)(____)(____)(____/                 *
*                                                                           *
*                                                                           *
*                                                                           *
*   This file is part of Circled                                            *
*                                                                           *
*   Copyright (C) 2011  Fabien LOISON                                       *
*                                                                           *
*   Circled is free software: you can redistribute it and/or modify         *
*   it under the terms of the GNU General Public License as published by    *
*   the Free Software Foundation, either version 3 of the License, or       *
*   (at your option) any later version.                                     *
*                                                                           *
*   This program is distributed in the hope that it will be useful,         *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*   GNU General Public License for more details.                            *
*                                                                           *
*   You should have received a copy of the GNU General Public License       *
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
*                                                                           *
****************************************************************************/


/**
 * This file contains all the class/functions for drawing/managing a TUI.
 */


#include "_headers.h"


//
// Event
//


/**
 * The constructor of the Event class.
 *
 * Initialize an empty event.
 */
Event::Event()
{
	//Fill the buffer with 0
	int i;
	for (i=0 ; i<=7 ; i++)
	{
		buff[i] = 0;
	}
}

/**
 * The constructor of the Event class.
 *
 * Initialize an event from a buffer.
 */
Event::Event(char input_buff[7])
{
	//Copy the buffer
	int i;
	for (i=0 ; i<=7 ; i++)
	{
		buff[i] = input_buff[i];
	}
}

/**
 * Get the type of an event.
 *
 * Returns:
 *     EV_TYPE_NONE (not implemented), EV_TYPE_CHAR (a simple char),
 *     EV_TYPE_KEYBOARD (keyboard special key) or EV_TYPE_MOUSE (mouse event).
 */
int Event::get_type()
{
	if (buff[0] != 0 && buff[1] == 0)
	{
		return EV_TYPE_CHAR;
	}
	else if (buff[0] == '\e' && buff[1] == '[' && buff[2] == 'M')
	{
		return EV_TYPE_MOUSE;
	}

	return EV_TYPE_NONE;
}

/**
 * Get the value of the char for a EV_TYPE_CHAR event.
 *
 * Returns:
 *     The char.
 */
char Event::get_char()
{
	return buff[0];
}

/**
 * Get the mouse event.
 *
 * Returns:
 *     MOUSE_DOWN or MOUSE_UP.
 */
char Event::get_mouse_event()
{
	return (buff[3] & MOUSE_MASK);
}

/**
 * Get the x position of mouse event.
 *
 * Returns:
 *     The x position.
 */
int Event::get_mouse_pos_x()
{
	return int(buff[4] - 33);
}

/**
 * Get the y position of mouse event.
 *
 * Returns:
 *     The y position.
 */
int Event::get_mouse_pos_y()
{
	return int(buff[5] - 33);
}


//
// Terminal
//


/**
 * The constructor of the Terminal class.
 *
 * Initialize the terminal for making a TUI with mouse support.
 */
Terminal::Terminal() 
{
	// Backup the terminal setting
	tcgetattr(0, &stored_settings);

	//Set the terminal IO stream properties
	termios new_settings;
	new_settings = stored_settings;
	new_settings.c_lflag &= (~ECHO);    //No echo
	new_settings.c_lflag &= (~ICANON);  //No line buffering
	new_settings.c_cc[VMIN] = 0;        //Minimal number of byte to read
	new_settings.c_cc[VTIME] = 1;       //Timout
    tcsetattr(0, TCSANOW, &new_settings);

	cout << "\e[?1047h"; //Alternate screen buffer
	cout << "\e[?25l";   //Hide cursor
	cout << "\e[?1001h"; //Mouse tracking (all events)

	cout << flush; //Flush
}

/**
 * The destructor of the Terminal class.
 *
 * Reset the terminal for being usable in line mode.
 */
Terminal::~Terminal()
{
	set_bg_color(-1); //Reset color and formatting attributes

	cout << "\e[?1001l"; //No mouse tracking
	cout << "\e[?25h";   //Show cursor
	cout << "\e[?1047l"; //Normal screen buffer

	cout << flush; //Flush

	//Restore the terminal setting
	tcsetattr(0, TCSANOW, &stored_settings);
}

/**
 * Get the terminal width.
 *
 * Returns:
 *     The terminal with.
 */
int Terminal::get_width()
{
    struct winsize w;
    ioctl(0, TIOCGWINSZ, &w);
	return w.ws_col;
}

/**
 * Get the terminal height.
 *
 * Returns:
 *     The terminal height.
 */
int Terminal::get_height()
{
    struct winsize w;
    ioctl(0, TIOCGWINSZ, &w);
	return w.ws_row;
}

/**
 * Set the foreground color.
 *
 * This color will be used for all next draw on the terminal.
 *
 * Args:
 *     color -- The color number (0..255).
 */
void Terminal::set_fg_color(int color)
{
	if (color < 0 ||  color > 255)
	{
		cout << "\e[0m" << flush;
	}
	else
	{
		cout << "\e[38;5;" << color << "m" << flush;
	}
}

/**
 * Set the background color.
 *
 * This color will be used for all next draw on the terminal.
 *
 * Args:
 *     color -- The color number (0..255).
 */
void Terminal::set_bg_color(int color)
{
	if (color < 0 || color > 255)
	{
		cout << "\e[0m" << flush;
	}
	else
	{
		cout << "\e[48;5;" << color << "m" << flush;
	}
}

/**
 * Move the cursor at the specified position.
 *
 * Args:
 *     x, y -- The coordinates of the new cursor position.
 */
void Terminal::move_xy(int x, int y)
{
	cout << "\e[" << y + 1 << ";" << x + 1 << "H" << flush;
}

/**
 * Draw a rectangle on the terminal.
 *
 * Args:
 *     x1, y1 -- The start coodrinate of the rectangle.
 *     x2, y2 -- The end coodrinate of the rectangle.
 */
void Terminal::draw_rectangle(int x1, int y1, int x2, int y2)
{
	string rectline = "";
	int i;

	for (i=x1 ; i<x2 ; i++)
	{
		rectline += " ";
	}

	for (i=y1 ; i<y2 ; i++)
	{
		move_xy(x1, i);
		cout << rectline << flush;
	}

	cout << flush; //Flush
}

/**
 * Draw a rectangle with a text inside.
 *
 * Args:
 *     x1, y1 -- The start coodrinate of the rectangle.
 *     x2, y2 -- The end coodrinate of the rectangle.
 *     text -- The text.
 */
void Terminal::draw_text_rectangle(int x1, int y1, int x2, int y2, char *text)
{
	draw_rectangle(x1, y1, x2, y2);

	string str(text);
	int x = x1 + (x2 - x1 - str.length()) / 2;
	int y = (y1 + y2) / 2;
	move_xy(x, y);

	cout << str << flush;
}

/**
 * Print a text at the specified position.
 *
 * Args:
 *     x, y -- The position where the text will be printed.
 *     str -- The text to print.
 */
void Terminal::print_xy(int x, int y, char* text)
{
	move_xy(x, y);
	cout << text << flush;
}

/**
 * Set the title of the terminal emulator's window.
 *
 * Args:
 *     title -- The new window title.
 */
void Terminal::set_title(char* title)
{
	cout << "\e]0;" << title << "\a" << flush;
}

/**
 * Wait for any event.
 *
 * Read the stdin stream until an event happen and then return the event.
 *
 * Returns:
 *     An Event instance containing all the information about the event.
 */
Event Terminal::get_event()
{
	char buff[7] = {0, 0, 0, 0, 0, 0, 0};

	do
	{
		read(0, &buff[0], 1);
		if (buff[0] == '\e')
		{
			read(0, &buff[1], 1);
			if (buff[1] == '[')
			{
				read(0, &buff[2], 1);
				if (buff[2] == 'M')
				{
					read(0, &buff[3], 1);
					read(0, &buff[4], 1);
					read(0, &buff[5], 1);
				}
			}
		}
	} while (buff[0] == 0);

	return Event(buff);
}


//
// MouseEventBox
//


/**
 * The constructor of the MouseEventBox class.
 *
 * Initialize the MouseEventBox with the given coordinates.
 *
 * Args:
 *     x1, y1 -- The start coodrinate of the box.
 *     x2, y2 -- The end coodrinate of the box.
 */
MouseEventBox::MouseEventBox(int x1, int y1, int x2, int y2)
{
	coord[0] = x1;
	coord[1] = y1;
	coord[2] = x2;
	coord[3] = y2;
}

/**
 * Check if there is a collide between the given coordinate and the event box.
 *
 * Args:
 *     ev -- The Event intance.
 *
 * Returns:
 *     true if there is a collide, false else.
 */
bool MouseEventBox::check_collide(Event *ev)
{
	if (ev->get_type() == EV_TYPE_MOUSE &&
	    ev->get_mouse_pos_x() >= coord[0] && ev->get_mouse_pos_x() < coord[2] &&
	    ev->get_mouse_pos_y() >= coord[1] && ev->get_mouse_pos_y() < coord[3])
	{
		return true;
	}
	else
	{
		return false;
	}
}

/**
 * Get the x potition (relative to the MouseEventPox position).
 *
 * Args:
 *     ev -- The Event intance.
 *
 * Returns:
 *     The relative x position.
 */
int MouseEventBox::get_relative_x(Event *ev)
{
	return ev->get_mouse_pos_x() - coord[0];
}

/**
 * Get the y potition (relative to the MouseEventPox position).
 *
 * Args:
 *     ev -- The Event intance.
 *
 * Returns:
 *     The relative y position.
 */
int MouseEventBox::get_relative_y(Event *ev)
{
	return ev->get_mouse_pos_y() - coord[1];
}
